"""
Author: Miroslav Bártík
Name of the program: Lloydova 15
Version: 1.0
date: 18.11.2024
description: This is a knowledge-base module that process and interacts with the database (in this case, the searching
             algorithm is implemented here along with other functionalities of the application)
"""

import heapq
import time
from Data_base import Puzzle


def a_star_solve(puzzle):
    """
    Solves the given sliding puzzle using the A* search algorithm.

    :param puzzle: Puzzle object representing the initial state of the puzzle.
    :return: Tuple containing the path to solve the puzzle, number of expanded nodes, and solving time.
    """
    start_time = time.time()  # Start the timer for solving time calculation
    start = puzzle.copy()  # Create a copy of the puzzle to preserve the original state
    start_state = (start.manhattan_distance(), 0, start, [])  # (heuristic, cost, puzzle, path)
    queue = []  # Priority queue to store puzzle states for A* exploration
    heapq.heappush(queue, start_state)  # Add the initial state to the queue
    visited = set()  # Set to track visited puzzle states
    num_expanded_nodes = 0  # Counter for the number of nodes expanded

    # Process the states in the queue until it's empty or the solution is found
    while queue:
        heuristic, cost, current_puzzle, path = heapq.heappop(queue)   # Get the state with the lowest heuristic

        # If the puzzle is solved, return the path of moves
        if is_solved(current_puzzle):
            solve_time = time.time() - start_time  # Calculate solve time
            return path, num_expanded_nodes, solve_time

        # Mark the current state as visited when it's being expanded
        board_tuple = tuple(tuple(row) for row in current_puzzle.grid)   # Convert the grid to a tuple for hashing
        if board_tuple in visited:
            continue
        visited.add(board_tuple)

        # Increment the expanded nodes counter
        num_expanded_nodes += 1

        # Generate all possible moves from the current state
        for move in get_neighbors(current_puzzle):
            new_puzzle = current_puzzle.copy()  # Create a copy of the puzzle for the move
            new_puzzle.move(move)  # Apply the move

            # Convert the new grid to a tuple for hashing
            new_board = tuple(tuple(row) for row in new_puzzle.grid)
            if new_board not in visited:
                new_path = path + [move]  # Add the move to the path
                new_cost = cost + 1  # Increment the cost
                heuristic = new_puzzle.manhattan_distance() + new_cost  # Calculate the heuristic
                heapq.heappush(queue, (heuristic, new_cost, new_puzzle, new_path))  # Add the new state to the queue

    return None  # Return None if no solution is found


def is_solvable(grid):
    """
    Checks if the given puzzle grid is solvable.

    :param grid: 2D list representing the puzzle grid.
    :return: True if the grid is solvable, False otherwise.
    """
    n = len(grid)  # Size of the grid (n x n)
    expected_tiles = set(range(n * n))  # Expected set of tiles (0 to n^2 - 1)
    actual_tiles = {tile for row in grid for tile in row}   # Flatten grid and create a set of tiles

    if actual_tiles != expected_tiles:
        return False  # Grid is invalid if it doesn't have exactly the expected tiles

    # Flatten the grid into a list for inversion count, ignoring the empty space (0)
    inversions = 0
    grid_flat = [tile for row in grid for tile in row if tile != 0]

    # Count inversions in the flattened grid
    for i in range(len(grid_flat)):
        for j in range(i + 1, len(grid_flat)):
            if grid_flat[i] > grid_flat[j]:
                inversions += 1

    # Find the row of the empty space (0), counting from the bottom
    empty_row = None
    for row_idx, row in enumerate(grid):
        if 0 in row:
            empty_row = n - row_idx  # Row position from the bottom
            break

    # Solvability check
    if n % 2 != 0:
        # For odd grid sizes, solvable if inversions are even
        return inversions % 2 == 0
    else:
        # For even grid sizes, solvable if:
        # (inversions even and empty on odd row) OR (inversions odd and empty on even row)
        return (inversions % 2 == 0) == (empty_row % 2 != 0)


def is_solved(puzzle: Puzzle):
    """
    Checks if the given puzzle is in the solved state.

    :param puzzle: Puzzle object to check.
    :return: True if the puzzle is solved, False otherwise.
    """
    n = puzzle.n  # Size of the grid (n x n)
    for i in range(n):
        for j in range(n):
            # Check if each tile is in its correct position
            if puzzle.grid[i][j] != (i * n + j + 1) % (n * n):
                return False  # Return False if any tile is out of place
    return True  # Return True if all tiles are in the correct position


def get_neighbors(puzzle: Puzzle):
    """
    Returns a list of valid moves for the empty tile in the puzzle.

    :param puzzle: Puzzle object representing the current state.
    :return: List of valid moves ("up", "down", "left", "right").
    """
    empty_x, empty_y = puzzle.empty_pos  # Get the position of the empty tile
    neighbors = []  # List to store valid moves

    # Add valid moves based on the empty tile's position
    if empty_x > 0:             # move up
        neighbors.append("up")
    if empty_x < puzzle.n - 1:  # move down
        neighbors.append("down")
    if empty_y > 0:             # move left
        neighbors.append("left")
    if empty_y < puzzle.n - 1:  # move right
        neighbors.append("right")

    return neighbors  # Return the list of valid moves


