"""
This module provides explanations regarding flights, including their gate assignments,
reasons for flight cancellations, and other related details. It interacts with the
facts and data to generate human-readable explanations about flight status.

Author:
    Václav Hajšman (hajsman2@student-zcu.cz)
Date:
    7.11.2024
Version:
    1.0
"""
from DataStructures.Flight import Flight
from DataStructures.Gate import Gate
from FactBase import Facts
from KnowledgeBase import Knowledge


class ExplanationModule:
    """
    A module responsible for generating explanations regarding flights.

    It provides methods to find a flight, explain the flight's status, and retrieve
    relevant information about the flight and its assigned gate (or the reason for its absence).

    Attributes:
        data (Knowledge): The data containing airport and flight-related information.
        facts (Facts): The facts that provide real-time information about gate occupations and cancellations.
    """

    def __init__(self):
        """
        Initializes the ExplanationModule with no data or facts.
        Sets the `data` and `facts` attributes to None.
        """
        self.knowledge = None
        self.facts = None

    def setKnowledge(self, knowledge: Knowledge):
        """
        Sets the data for the ExplanationModule.

        Parameters:
            data (Knowledge): The data containing the airport and flight information to be used by the module.
        """
        self.knowledge = knowledge

    def setFacts(self, facts: Facts):
        """
        Sets the facts for the ExplanationModule.

        Parameters:
            facts (Facts): The facts containing gate occupation details and flight cancellations.
        """
        self.facts = facts

    def findFlight(self, flightId: int, gateOccupationList: dict, canceledFlights: list) -> tuple:
        """
        Finds a flight based on its flight ID.

        This method checks if the flight is assigned to a gate or if it's parked in the
        free parking area or if it's canceled.

        Parameters:
            flightId (int): The unique identifier of the flight.

        Returns:
            tuple: A tuple containing the found Flight object and the gate (or "freePark" if parked).
                   If no flight is found, returns (None, None).
        """
        for gate in self.facts.airport.gates:
            for flight in gateOccupationList[gate.name]:
                if flight.flightId == flightId:
                    return flight, gate

        for flight in gateOccupationList["freePark"]:
            if flight.flightId == flightId:
                return flight, "freePark"

        for flight in canceledFlights:
            if flight.flightId == flightId:
                return flight, None

        return None, None

    def explainFlight(self, flightId: int, gateOccupationList: dict, canceledFlights: list) -> list[str]:
        """
        Provides an explanation for a flight, including its status, gate assignment,
        and other relevant details.

        Parameters:
            flightId (int): The unique identifier of the flight to explain.

        Returns:
            list[str]: A list of strings that explain the flight's status.
        """
        explanations = []

        flight, gate = self.findFlight(flightId, gateOccupationList, canceledFlights)
        if not flight:
            explanations.append("Let s daným indexem nebyl nalezen!")
            return explanations

        explanations.append(flight.__str__())
        explanations = self.getFlightInfo(flight, gate, explanations, gateOccupationList)

        return explanations

    def getFlightInfo(self, flight: Flight, gate: Gate, explanations: list[str], gateOccupationList: dict) -> list[str]:
        """
        Retrieves additional information about the flight's gate assignment and explains
        why it could not use a certain gate or why it was canceled.

        Parameters:
            flight (Flight): The Flight object for which to retrieve information.
            gate (Gate or str): The gate assigned to the flight or "freePark" if parked.
            explanations (list[str]): The list of explanations to append the results to.

        Returns:
            list[str]: A list of strings containing additional explanations about the flight's status.
        """
        if not gate:
            explanations.append("Let s daným indexem byl zrušen pro nedostatek kapacity letiště!")
        elif gate == "freePark":
            explanations.append("Let s daným indexem byl odbaven na volné ploše!")
        else:
            explanations.append(f"Tento let obsadil stojánek: {gate.__str__()}!")

        # Check for gate occupation conflicts
        for possibleGate in flight.aircraft.gates:
            if gate and gate != "freePark" and possibleGate.name == gate.name:
                break
            if self.knowledge.isOccupied(possibleGate, flight.arrival, flight.departure, gateOccupationList):
                currentFlights = self.knowledge.getOccupiedFlightTotal(possibleGate, flight.arrival, flight.departure, gateOccupationList)
                explanation = f"Let nemohl použít stojánek: {possibleGate.__str__()}, protože byl obsazen lety: "
                for currentFlight in currentFlights:
                    explanation += f"{currentFlight.__str__()}, "
                explanation = explanation[:-2] + "."
                explanations.append(explanation)

        if not flight.aircraft.freePark and not gate:
            explanations.append("Let nemohl ani využít volné stání, protože jeho typ to nedovoluje.")

        return explanations
