"""
This module generates solutions for flight scheduling by evaluating each flight's fit
within the airport's capacity and handling any conflicts or issues. It interacts with
the facts and knowledge to process the flight table and generate the necessary adjustments.

Author:
    Václav Hajšman (hajsman2@student-zcu.cz)
Date:
    10.12.2024
Version:
    2.0
"""
from datetime import datetime

from DataStructures.Gate import Gate
from FactBase import Facts
from KnowledgeBase import Knowledge


class SolutionGenerator:
    """
    A module responsible for generating solutions related to flight scheduling.

    It evaluates each flight against the available airport data and facts, determining
    if any flight does not fit the current scheduling constraints. If a flight doesn't
    fit, the module handles the conflict by taking appropriate actions.

    Attributes:
        knowledge (Knowledge): The class containing all crucial functions for the computing logic.
        facts (Facts): The facts containing all the user given data.
        gateOccupationList (dict): The dictionary with all gates with described occupation by flights.
        canceledFlights (list): The list of all canceled flights.
    """

    def __init__(self):
        """
        Initializes the SolutionGenerator with no data or facts.
        Sets the `data` and `facts` attributes to None.
        Initializes the empty lists and dictionary.
        """
        self.knowledge = None
        self.facts = None
        self.gateOccupationList = {}
        self.canceledFlights = []

    def setKnowledge(self, knowledge: Knowledge):
        """
        Sets the data for the SolutionGenerator.

        Parameters:
            knowledge (Knowledge): The class containing all crucial functions for the computing logic.
        """
        self.knowledge = knowledge

    def setFacts(self, facts: Facts):
        """
        Sets the facts for the SolutionGenerator.

        Parameters:
            facts (Facts): The facts containing all the user given data.
        """
        self.facts = facts

    def initGateOccupationList(self, gates: list[Gate]):
        """
        Initializes the gate occupation list with given gates.

        Parameters:
            gates (list[Gate]): List of gates to initialize in the occupation list.
        """
        for gate in gates:
            self.gateOccupationList[gate.name] = []

        self.gateOccupationList["freePark"] = []

    def calculate(self):
        """
        Processes the flight table sorted by the occupation time and calculates the solution
        for any flights that do not fit within the constraints.

        For each flight in the flight table, the method checks if the flight can be scheduled
        based on the available gates and airport capacity. If a flight does not fit, it invokes
        the appropriate conflict resolution method to handle the situation.
        """

        # sort flight table based on the occupation duration
        self.facts.flightTable.sort(key=lambda f: (
                datetime.combine(datetime.today(), f.departure) -
                datetime.combine(datetime.today(), f.arrival)
        ).total_seconds())

        # calculate optimal gate occupation
        for flight in self.facts.flightTable:
            if not self.knowledge.fitFlight(flight, self.gateOccupationList):
                self.knowledge.handleNotFittingFlight(flight, self.gateOccupationList, self.canceledFlights)

        # sort all gates based on the arrival time
        for gate in self.facts.airport.gates:
            self.gateOccupationList[gate.name] = sorted(self.gateOccupationList[gate.name],
                                                        key=lambda f: f.arrival)

        self.gateOccupationList["freePark"] = sorted(self.gateOccupationList["freePark"],
                                                     key=lambda f: f.arrival)

        self.facts.canceledFlights = sorted(self.canceledFlights,
                                            key=lambda f: f.arrival)
