"""
This module defines the class @UserInterface, which serves as the communication
module with the user who controls the knowledge base system programme.
The class handles the tasks of printing the data to the command line
and to the output file. Another important task for the class is to retrieve
the user given data mainly from the command line. The module also has function
which are essential for the file reading and parsing.

Author:
    Václav Hajšman (hajsman2@student-zcu.cz)
Date:
    10.12.2024
Version:
    2.0
"""
import sys
from datetime import time

from DataStructures.Aircraft import Aircraft
from DataStructures.Flight import Flight
from DataStructures.Airport import Airport
from DataStructures.Gate import Gate
from FactBase import Facts

import os
from typing import List, Tuple


def loadFile(filePath: str) -> List[str]:
    """
    Loads the contents of a file and splits it into a list of strings, one for each line.

    Parameters:
        filePath : str
            The path to the file to be loaded.

    Returns:
        List[str]
            A list of strings where each element is a line from the file.

    Raises:
        FileNotFoundError:
            If the specified file does not exist, prints an error message.
    """
    if not os.path.exists(filePath):
        print(f"Vstupní soubor '{filePath}' neexistuje!")
        sys.exit(1)

    with open(filePath, 'r', encoding='utf-8') as file:
        text = file.read().split("\n")
    return text


def parseConfigFile(inputText: List[str]) -> Tuple[Airport, List[Aircraft]]:
    """
    Parses configuration data for the airport and aircraft from the input text.

    Parameters:
        inputText : List[str]
            A list of strings representing the lines read from the configuration file.

    Returns:
        Tuple[Airport, List[Aircraft]]
            A tuple containing:
                - `Airport`: The airport configuration with gates and free parking count.
                - `List[Aircraft]`: A list of `Aircraft` objects representing the aircraft in the configuration.
    """
    aircraftList = []
    gatesList = []
    freeCount = 0

    readID = 0
    lineId = 0

    while lineId < len(inputText):
        line = inputText[lineId].strip()

        if line == "":
            lineId += 1
            continue

        if line == "aircraft:":
            readID = 1
            lineId += 1
            continue
        elif line == "airport:":
            readID = 2
            lineId += 1
            continue

        if readID == 1:
            parts = line.split(';')
            name = parts[0]
            gateNames = parts[1:]

            freePark = not gateNames[-1] == "x"
            if not freePark:
                gateNames.pop()

            gates = [Gate(gateName) for gateName in gateNames]
            aircraftList.append(Aircraft(name, gates, freePark))

        elif readID == 2:
            if line.startswith("free"):
                freeCount = int(line.split(" ")[1])
            else:
                gatesList.append(Gate(line))

        lineId += 1

    return Airport(gatesList, freeCount), aircraftList


def parse_time(time_str: str) -> time:
    """
    Parses a time string in the format 'HH:MM' and returns a `time` object.

    Parameters:
        time_str : str
            The time string to be parsed, expected to be in the format 'HH:MM'.

    Returns:
        time
            A `time` object representing the parsed time.
    """
    hours, minutes = map(int, time_str.split(":"))
    return time(hours, minutes)


def getAircraft(aircraftName: str, aircraftList: list[Aircraft]) -> Aircraft | None:
    """
    Finds and returns an `Aircraft` object from the list of aircraft by name.

    Parameters:
        aircraftName : str
            The name of the aircraft to be found.
        aircraftList : list[Aircraft]
            A list of `Aircraft` objects.

    Returns:
        Aircraft | None
            The `Aircraft` object if found, or `None` if no matching aircraft is found.
    """
    for aircraft in aircraftList:
        if aircraft.name == aircraftName:
            return aircraft

    return None


def parseFlightTableFile(inputText: List[str], aircraftList: list[Aircraft]) -> List[Flight]:
    """
    Parses flight table data from the input text and returns a sorted list of `Flight` objects.

    Parameters:
        inputText : List[str]
            A list of strings representing the lines read from the flight table file.
        aircraftList : list[Aircraft]
            A list of `Aircraft` objects to match flight data.

    Returns:
        List[Flight]
            A sorted list of `Flight` objects parsed from the flight table data, sorted by arrival time.
    """
    flightList = []
    lineId = 0

    while lineId < len(inputText):
        aircraftName = inputText[lineId].strip()
        lineId += 1

        if not aircraftName or lineId >= len(inputText):
            continue

        aircraft = getAircraft(aircraftName, aircraftList)

        if not aircraft:
            continue

        while lineId < len(inputText) and ";" in inputText[lineId]:
            times = inputText[lineId].strip().split(";")
            arrival = parse_time(times[0])
            departure = parse_time(times[1])
            flight = Flight(aircraft, arrival, departure)
            if arrival < departure:
                flightList.append(flight)
            lineId += 1

    sorted_flights = sorted(flightList, key=lambda f: f.arrival)
    return sorted_flights


class UserInterface:
    """
    A class to handle the user interface for the flight scheduling application.
    This class interacts with the user to load data, display information, and collect inputs.

    Attributes:
        outputFile (File): Provides the output file for the writing.
    """

    def __init__(self):
        """
        Initializes the UserInterface instance and opens the output file `výsledek.txt`
        for writing results.
        """
        self.outputFile = open("výsledek.txt", 'w')

    def myPrint(self, text: str):
        """
        Prints the provided text to both the console and the output file.

        Parameters:
        text (str): The text to print.
        """
        self.outputFile.write(text + '\n')
        print(text)

    def validateProgrammeArgs(self):
        """
        Validates if the programme arguments are in correct format.
        First argument must be configuration file with definition
        of the airport and definition of the aircraft. Second argument
        is the flight table. It does not have to be specified if users
        chooses to randomly generate flight table.
        """
        if len(sys.argv) < 2:
            self.myPrint("Programe nemá zadané potřebné parametry!")
            self.myPrint("1. parametr (povinný) - cesta k inicializačnímu souboru")
            self.myPrint("2. parametr (nepovinný) - cesta k souboru s letovými daty (bez tohoto souboru lze použít jen generování dat)")
            input("STOP")
            sys.exit(1)

    def printProgrammeInfo(self):
        """
        Displays a welcome message and the instructions for interacting with the program.
        """
        self.myPrint("----------------------------------------------------------------------------------------------------")
        self.myPrint("---------------------- Vítejte v aplikaci plánovače odbavení letů na letišti! ----------------------")
        self.myPrint("---- Tato aplikace Vám poskytne optimální rozložení obsazení stojánků a volné plochy na letišti ----")
        self.myPrint("----------------------------------------------------------------------------------------------------")
        self.myPrint("")
        self.myPrint("Ovládání: ")
        self.myPrint("  Uživatel má možnost ovlivnit pouze s jakými daty program pracuje")
        self.myPrint("  Program provede výpočet obsazení stojánků sám")
        self.myPrint("  Zároveň program pracuje s množinou dat popisujících letadla a letiště ze souboru daného prvním argumentem programu")
        self.myPrint("  Tato data lze změnit v souboru config/init.txt")
        self.myPrint("  Uživatel má možnost rozhodnout zdali program využije úkázková data ze souboru config/table.txt")
        self.myPrint("  Nebo lze vygenerovat svoje vlastní náhodná data")
        self.myPrint("  Po provedení výpočtu může uživatel požádat o vysvětlení daného letu zadáním jeho indexu")
        self.myPrint("  Všechny výpisy jsou zároveň prováděny do souboru výsledek.txt a po ukončení programu jsou zde k nalezení")
        self.myPrint("")

    def getConfigFileData(self) -> tuple[Airport, list[Aircraft]]:
        """
        Loads the configuration data from the file specified by the first command-line argument
        and parses it to return an `Airport` object and a list of `Aircraft` objects.

        Returns:
        tuple: A tuple containing:
            - `Airport`: The parsed airport data.
            - `List[Aircraft]`: The list of parsed aircraft.
        """
        inputText = loadFile(sys.argv[1])
        return parseConfigFile(inputText)

    def flightTableChoice(self) -> int:
        """
        Prompts the user to choose between generating flight data or using the default flight data from a file.

        Returns:
        int:
            - 0 if the user selects to use the default data file.
            - A number between 1 and 100 if the user opts to generate flight data.
        """
        userInput = 0
        while userInput != "1" and userInput != "2":
            userInput = input(
                "Přejete si vygenerovat data pro lety nebo využít základní data? (1 - generovat, 2 - soubor): ")

        if userInput == "1":
            userInput = "250"
            while not userInput.isdigit() or not 0 < int(userInput) < 101:
                userInput = input(
                    "Proběhne generování! Zadejte počet letů, které chcete vygenerovat (celé číslo 1-100): ")
            return int(userInput)

        return 0

    def getFlightTableData(self, aircraftList: list[Aircraft]) -> list[Flight]:
        """
        Loads the flight table data from the file specified by the second command-line argument
        and parses it using the provided `aircraftList`.

        Parameters:
        aircraftList (list[Aircraft]): A list of aircraft objects used to parse the flight data.

        Returns:
        list: A sorted list of `Flight` objects parsed from the flight table data.
        """
        if len(sys.argv) < 3:
            self.myPrint("Programe nemá zadané potřebné parametry!")
            self.myPrint("Zvolili jste zadání letů pomocí vstupního souboru, avšak tento soubor není v programových parametrech!")
            input("STOP")
            sys.exit(1)

        inputText = loadFile(sys.argv[2])
        return parseFlightTableFile(inputText, aircraftList)

    def printLoadedData(self, facts: Facts):
        """
        Prints the loaded data (airport, aircraft, and flight table) to the console and the output file.

        Parameters:
        data (Data): The `Data` object containing the loaded airport, aircraft, and flight data.
        """
        self.myPrint("")
        self.myPrint("Načtená data:")
        self.myPrint("")
        self.myPrint("Letiště: ")
        self.myPrint("Počet volných míst: " + str(facts.airport.freeParkCount))

        for gate in facts.airport.gates:
            self.myPrint(gate.__str__())

        self.myPrint("")
        self.myPrint("Letadla: ")

        for aircraft in facts.aircraftList:
            self.myPrint(aircraft.__str__())

        self.myPrint("")
        self.myPrint("Plánované lety: ")

        for flight in facts.flightTable:
            self.myPrint(flight.__str__())

    def printOutput(self, occupationList: dict[str], canceledFlights: list[Flight]):
        """
        Prints the planned gate occupation and canceled flights.

        Parameters:
        occupationList (dict[str]): A dictionary mapping gate names to the list of flights occupying those gates.
        canceledFlights (list[Flight]): A list of canceled flights.
        """
        self.myPrint("")
        self.myPrint("Výpis plánovaného rozmístění na branách:")
        self.myPrint("")

        for gate_name, flights in occupationList.items():
            if gate_name != "freePark":
                self.myPrint("Stojánek: " + gate_name)
                for flight in flights:
                    self.myPrint("   " + flight.__str__())
                self.myPrint("")
            else:
                self.myPrint("Volné stání:")
                for flight in flights:
                    self.myPrint("   " + flight.__str__())
                self.myPrint("")

        self.myPrint("Zrušené lety:")

        if len(canceledFlights) <= 0:
            self.myPrint("   Žádné lety nebyly zrušeny")
        else:
            for flight in canceledFlights:
                self.myPrint("   " + flight.__str__())

    def readFlightId(self) -> str | None:
        """
        Prompts the user for a flight ID for which they want an explanation.
        Returns the flight ID as an integer or None if the user enters `"konec"` to end the program.

        Returns:
            - An string representing the flight ID if valid input is provided.
            - `None` if the user enters `"konec"`.
        """
        self.myPrint("")
        flightId = input("Zadejte číslo letu, pro který chcete dostat vysvětlení (celé číslo) nebo ukončete program (\"konec\"): ")

        while not flightId.isdigit() and not flightId == "konec":
            flightId = input("Zadejte číslo letu, pro který chcete dostat vysvětlení (celé číslo) nebo ukončete program (\"konec\"): ")

        if flightId == "konec":
            self.outputFile.close()
            return None

        return flightId

    def printExplanation(self, explanations: list[str]):
        """
        Prints the explanations for a particular flight to the console and the output file.

        Parameters:
        explanations (list[str]): A list of explanation strings for the specified flight.
        """
        self.myPrint("")

        for explanation in explanations:
            self.myPrint(explanation)

